from hypothesis import given, settings
from sys import path
from generation import acc_palindromes, rej_palindromes

path.append("../src/ib110hw")
from turing.mtm import MTM
from turing.tape import Direction
from turing.utils import import_mtm


PALINDROME_MACHINE: MTM = MTM(
    states={"init", "goToEnd", "goToStart", "check", "accept", "reject"},
    initial_state="init",
    input_alphabet={"a", "b"},
    acc_state="accept",
    rej_state="reject",
    transitions={
        "init": {(">", ""): ("copy", (">", ""), (Direction.RIGHT, Direction.STAY))},
        "copy": {
            ("a", ""): ("copy", ("a", "a"), (Direction.RIGHT, Direction.RIGHT)),
            ("b", ""): ("copy", ("b", "b"), (Direction.RIGHT, Direction.RIGHT)),
            ("", ""): ("goToStart", ("", ""), (Direction.LEFT, Direction.STAY)),
        },
        "goToStart": {
            ("a", ""): ("goToStart", ("a", ""), (Direction.LEFT, Direction.STAY)),
            ("b", ""): ("goToStart", ("b", ""), (Direction.LEFT, Direction.STAY)),
            (">", ""): ("check", (">", ""), (Direction.RIGHT, Direction.LEFT)),
        },
        "check": {
            ("a", "a"): ("check", ("a", "a"), (Direction.RIGHT, Direction.LEFT)),
            ("b", "b"): ("check", ("b", "b"), (Direction.RIGHT, Direction.LEFT)),
            ("", ""): ("accept", ("", ""), (Direction.STAY, Direction.STAY)),
            ("a", "b"): ("reject", ("a", "b"), (Direction.STAY, Direction.STAY)),
            ("b", "a"): ("reject", ("b", "a"), (Direction.STAY, Direction.STAY)),
        },
    },
)

PALINDROME_MACHINE.max_steps = 10000


@given(acc_palindromes())
def test_simulate_acc(input_str: str):
    PALINDROME_MACHINE.write_to_tape(input_str)
    assert PALINDROME_MACHINE.simulate(to_console=False)
    PALINDROME_MACHINE.clear_tapes()


@given(rej_palindromes())
def test_simulate_rej(input_str: str):
    PALINDROME_MACHINE.write_to_tape(input_str)
    assert not PALINDROME_MACHINE.simulate(to_console=False)
    PALINDROME_MACHINE.clear_tapes()


@given(acc_palindromes())
@settings(deadline=None)
def test_import_acc(input_str: str):
    machine = import_mtm("./inputs/mtm_input")
    machine.max_steps = 10000
    machine.write_to_tape(input_str)
    assert machine.simulate(to_console=False)

@given(rej_palindromes())
@settings(deadline=None)
def test_import_rej(input_str: str):
    machine = import_mtm("./inputs/mtm_input")
    machine.max_steps = 10000
    machine.write_to_tape(input_str)
    assert not machine.simulate(to_console=False)

def test_import_invalid_transitions_tapes_missing():
    assert not import_mtm("./inputs/mtm_input_invalid_tapes_missing")

def test_import_invalid_transitions_tapes_number():
    assert not import_mtm("./inputs/mtm_input_invalid_tapes_number")

def test_import_invalid_transitions_arrow():
    assert not import_mtm("./inputs/mtm_input_invalid_transitions_arrow")

def test_import_invalid_transitions_direction():
    assert not import_mtm("./inputs/mtm_input_invalid_transitions_direction")

def test_import_invalid_transitions_state():
    assert not import_mtm("./inputs/mtm_input_invalid_transitions_state")

def test_import_invalid_transitions_next():
    assert not import_mtm("./inputs/mtm_input_invalid_transitions_next")

def test_import_invalid_transitions_write():
    assert not import_mtm("./inputs/mtm_input_invalid_transitions_write")



from sys import path
from hypothesis import given, assume
from hypothesis.strategies import integers, sets, characters, composite, DrawFn
from tests.generation import r_dfa
from random import choice

path.append("../src/ib110hw")

from automaton.dfa import DFA


@composite
def r_test_dfa(
    draw: DrawFn,
    min_deg=integers(min_value=1, max_value=10),
    max_deg=integers(min_value=1, max_value=10),
    min_states=integers(min_value=3, max_value=10),
    max_states=integers(min_value=3, max_value=10),
    min_fin_states=integers(min_value=1, max_value=10),
    max_fin_states=integers(min_value=2, max_value=10),
    alphabet=sets(characters(), min_size=2, max_size=20),
):
    max_deg = draw(max_deg)
    min_deg = min(draw(min_deg), max_deg)
    max_states = draw(max_states)
    min_states = min(draw(min_states), max_states)
    max_fin_states = draw(max_fin_states)
    min_fin_states = min(draw(min_fin_states), max_fin_states)

    r_automaton = r_dfa(
        min_deg,
        max_deg,
        min_states,
        max_states,
        min_fin_states,
        max_fin_states,
        draw(alphabet),
    )

    diff = r_automaton.states.difference(r_automaton.final_states)
    assume(diff and diff != {r_automaton.initial_state})

    return r_automaton


@given(r_test_dfa())
def test_add_state(automaton: DFA) -> None:
    assert automaton.add_state("test_state_1")
    # cannot add the same state twice
    assert not automaton.add_state("test_state_1")
    assert not automaton.add_state("test_state_1", True)

    assert "test_state_1" in automaton.states
    assert "test_state_1" not in automaton.final_states
    assert "test_state_1" not in automaton.transitions

    assert automaton.add_state("test_state_2", True)
    assert not automaton.add_state("test_state_2")
    assert "test_state_2" in automaton.states
    assert "test_state_2" in automaton.final_states


@given(r_test_dfa())
def test_remove_state(automaton: DFA):
    state_to_remove = choice(
        [s for s in automaton.states if s not in automaton.final_states]
    )

    assert automaton.remove_state(state_to_remove)
    # cannot remove the same state twice
    assert not automaton.remove_state(state_to_remove)
    assert state_to_remove not in automaton.states
    assert not automaton.remove_state("not_existent")

    state_to_remove = choice(list(automaton.final_states))

    assert automaton.remove_state(state_to_remove)
    assert state_to_remove not in automaton.states
    assert state_to_remove not in automaton.final_states
    assert not automaton.transitions.get(state_to_remove, None)


@given(r_test_dfa())
def test_add_transition(automaton: DFA) -> None:
    assert automaton.add_transition("test_state", "test_state", "a")
    assert "test_state" in automaton.transitions.keys()
    assert "a" in automaton.transitions["test_state"].keys()
    assert "test_state" == automaton.transitions["test_state"]["a"]
    assert not automaton.add_transition("test_state", "test_state", "a")


@given(r_test_dfa())
def test_set_transition(automaton: DFA) -> None:
    automaton.set_transition("s_from", "s_to", "symbol")
    assert automaton.get_transition("s_from", "symbol") == "s_to"

    automaton.set_transition("s_from", "s_to2", "symbol2")
    assert automaton.get_transition("s_from", "symbol2") == "s_to2"
    assert automaton.get_transition("s_from", "symbol") == "s_to"

    automaton.set_transition("s_from", "s_to2", "symbol")
    assert automaton.get_transition("s_from", "symbol") == "s_to2"


@given(r_test_dfa())
def test_remove_transition(automaton: DFA) -> None:
    automaton.add_state("test_state")
    automaton.alphabet.add("a")
    automaton.add_transition("test_state", automaton.initial_state, "a")

    assert not automaton.remove_transition("test_state", "not_exists")
    assert not automaton.remove_transition(
        "test_state", automaton.alphabet.difference(["a"]).pop()
    )
    assert not automaton.remove_transition("not_exists", "a")
    assert automaton.remove_transition("test_state", "a")
    assert not automaton.get_transition("test_state", "a")


def test_is_accepted() -> None:
    automaton: DFA = DFA(
        states={"s0", "s1", "s2", "s3", "s4"},
        alphabet={"a", "b", "c"},
        initial_state="s0",
        final_states={"s0", "s3"},
        transitions={
            "s0": {"a": "s0", "b": "s0", "c": "s1"},
            "s1": {"a": "s4", "b": "s2", "c": "s1"},
            "s2": {"a": "s3", "b": "s1", "c": "s1"},
            "s3": {"a": "s3", "b": "s3", "c": "s2"},
            "s4": {"a": "s4", "b": "s4", "c": "s4"},
        },
    )

    # words accepted in s0
    acc_s0 = ["", "a", "b", "ab", "aba", "abba"]
    # words accepted in s3 (starting from s1)
    acc_s3 = ["ba", "cbaab", "cbaca", "bbba", "bcba", "baccbab"]
    # words accepted in s0 and s1 (starting from s0)
    acc = [f"{w0}c{w1}" for w0 in acc_s0 for w1 in acc_s3] + acc_s0

    for acc_str in acc:
        assert automaton.is_accepted(acc_str)

    # words rejected in s1
    rej_s1 = ["c", "cc", "cbb", "cbc", "cbb", "cbbc"]
    # words rejected in s2 (starting from s1)
    rej_s2 = ["b", "bbc", "bac", "baac", "babc", "baccb"]
    # words rejected in s4
    rej_s4 = [f"{w}a" for w in rej_s1]
    # words rejected in s1, s2 and s4 (starting from s0)
    rej = [
        f"{w}{w4}"
        for w in [f"{w1}{w2}" for w1 in rej_s1 for w2 in rej_s2]
        for w4 in rej_s4
    ] + rej_s1

    for rej_str in rej:
        assert not automaton.is_accepted(rej_str)


def test_is_valid() -> None:
    automaton: DFA = DFA(
        states={"s0", "s1", "s2"},
        alphabet={"a", "b", "c"},
        initial_state="s0",
        final_states={"s1", "s2"},
        transitions={
            "s0": {"a": "s0", "b": "s2", "c": "s1"},
            "s1": {"a": "s1", "b": "s1", "c": "s2"},
            "s2": {"a": "s1", "b": "s0", "c": "s1"},
        },
    )

    assert automaton.is_valid()

    # invalid initial state
    automaton.initial_state = "s3"
    assert not automaton.is_valid()

    automaton.initial_state = "s0"
    assert automaton.is_valid()

    # invalid state in the transition function
    automaton.set_transition("s0", "s3", "c")
    assert not automaton.is_valid()

    automaton.set_transition("s0", "s1", "c")
    assert automaton.is_valid()

    # final states set is not subset of the states set
    automaton.final_states.update({"s3"})
    assert not automaton.is_valid()

    automaton.final_states.remove("s3")
    assert automaton.is_valid()

    # invalid character in the transition function
    automaton.add_transition("s0", "s1", "x")
    assert not automaton.is_valid()

    automaton.remove_transition("s0", "x")
    assert automaton.is_valid()

    # transition function is not total
    automaton.remove_transition("s0", "c")
    assert not automaton.is_valid()

    automaton.add_transition("s0", "s1", "c")
    assert automaton.is_valid()

    # empty states set
    automaton.states = set()
    assert not automaton.is_valid()



from hypothesis import given, settings
from sys import path
from generation import acc_palindromes, rej_palindromes

path.append("../src/ib110hw")
from turing.dtm import DTM
from turing.tape import Direction
from turing.utils import import_dtm
import pathlib

PALINDROME_MACHINE: DTM = DTM(
    states={
        "init",
        "mark",
        "gotoEndA",
        "checkA",
        "gotoEndB",
        "checkB",
        "accept",
        "reject",
    },
    input_alphabet={"a", "b"},
    acc_state="accept",
    rej_state="reject",
    initial_state="init",
    transitions={
        "init": {
            ">": ("mark", ">", Direction.RIGHT),
        },
        "mark": {
            "a": ("foundA", "X", Direction.RIGHT),
            "b": ("foundB", "X", Direction.RIGHT),
            "X": ("accept", "X", Direction.STAY),
            "": ("accept", "", Direction.STAY),
        },
        "foundA": {
            "a": ("foundA", "a", Direction.RIGHT),
            "b": ("foundA", "b", Direction.RIGHT),
            "X": ("checkA", "X", Direction.LEFT),
            "": ("checkA", "", Direction.LEFT),
        },
        "checkA": {
            "a": ("back", "X", Direction.LEFT),
            "b": ("reject", "b", Direction.STAY),
            "X": ("accept", "X", Direction.STAY),
        },
        "foundB": {
            "a": ("foundB", "a", Direction.RIGHT),
            "b": ("foundB", "b", Direction.RIGHT),
            "X": ("checkB", "X", Direction.LEFT),
            "": ("checkB", "", Direction.LEFT),
        },
        "checkB": {
            "a": ("reject", "a", Direction.STAY),
            "b": ("back", "X", Direction.LEFT),
            "X": ("accept", "X", Direction.STAY),
        },
        "back": {
            "a": ("back", "a", Direction.LEFT),
            "b": ("back", "b", Direction.LEFT),
            "X": ("mark", "X", Direction.RIGHT),
        },
    },
)

PALINDROME_MACHINE.max_steps = 10000


@given(acc_palindromes())
@settings(deadline=None)
def test_simulate_acc(input_str: str):
    PALINDROME_MACHINE.write_to_tape(input_str)
    assert PALINDROME_MACHINE.simulate(to_console=False)
    PALINDROME_MACHINE.clear_tape()


@given(rej_palindromes())
@settings(deadline=None)
def test_simulate_rej(input_str: str):
    PALINDROME_MACHINE.write_to_tape(input_str)
    assert not PALINDROME_MACHINE.simulate(to_console=False)
    PALINDROME_MACHINE.clear_tape()

@given(acc_palindromes())
@settings(deadline=None)
def test_import_acc(input_str: str):
    machine = import_dtm("inputs/dtm_input")
    machine.max_steps = 10000
    machine.write_to_tape(input_str)
    assert machine.simulate(to_console=False)

@given(rej_palindromes())
@settings(deadline=None)
def test_import_rej(input_str: str):
    machine = import_dtm("inputs/dtm_input")
    machine.max_steps = 10000
    machine.write_to_tape(input_str)
    assert not machine.simulate(to_console=False)

def test_import_invalid_init():
    assert not import_dtm("./inputs/dtm_input_invalid_init")

def test_import_invalid_acc():
    assert not import_dtm("./inputs/dtm_input_invalid_acc")

def test_import_invalid_rej():
    assert not import_dtm("./inputs/dtm_input_invalid_rej")

def test_import_invalid_alphabet():
    assert not import_dtm("./inputs/dtm_input_invalid_alphabet")

def test_import_invalid_transitions_arrow():
    assert not import_dtm("./inputs/dtm_input_invalid_transitions_arrow")

def test_import_invalid_transitions_direction():
    assert not import_dtm("./inputs/dtm_input_invalid_transitions_direction")

def test_import_invalid_transitions_state():
    assert not import_dtm("./inputs/dtm_input_invalid_transitions_state")

def test_import_invalid_transitions_next():
    assert not import_dtm("./inputs/dtm_input_invalid_transitions_next")


from typing import Set

MAX_STEPS_ERROR_MSG = """
Exceeded the maximum allowed steps. ({})
You change the default value by setting the 'max_steps' property of this automaton."""


class BaseTuringMachine:
    """Represents an abstract Turing machine class. This class cannot be instantiated."""

    def __new__(cls, *args, **kwargs):
        if cls is BaseTuringMachine:
            raise TypeError("Only DTM and MTM can be instantiated!")

        return object.__new__(cls)

    def __init__(
        self,
        states: Set[str],
        input_alphabet: Set[str],
        acc_state: str = "accept",
        rej_state: str = "reject",
        initial_state: str = "init",
        start_symbol: str = ">",
    ) -> None:
        self.states = states or set()
        self.input_alphabet = input_alphabet or set()
        self.acc_state = acc_state
        self.rej_state = rej_state
        self.initial_state = initial_state
        self.start_symbol = start_symbol

        # After exceeding max_steps value, the turing machine will be considered as looping.
        # Change this value for more complex scenarios.
        self.max_steps = 100


if __name__ == "__main__":
    pass



from os import name, system
from typing import IO, List, Tuple, Set, Optional, Callable
from .tape import Tape, Direction
from itertools import takewhile, dropwhile
from pynput import keyboard
import re


def clear_console() -> None:
    if name == "posix":
        system("clear")
    else:
        system("cls")


def close_file(file: IO) -> None:
    if file:
        file.close()


# inspired by https://stackoverflow.com/a/43106497
def get_step_direction() -> str:
    step_direction = None
    print("Use arrow keys [L|R] to go back or forward. Press ESC to exit.")

    def on_press(key):
        nonlocal step_direction
        try:
            if key.name in ["left", "right", "esc"]:
                step_direction = key.name
                return False
        except AttributeError:
            # ignore other keys
            pass

    listener = keyboard.Listener(on_press=on_press)
    listener.start()
    listener.join()
    return step_direction


def tape_to_md(tape: Tape, index: int = None) -> str:
    index_str = "" if index is None else f"<b>Tape {index}:</b>"
    tape_str = f"{tape}".replace("^", "<b>^</b>")
    return f"{index_str}\n{tape_str}\n"


def rule_to_md(rule_str: str) -> str:
    return f"######{rule_str.replace('->', '&rarr;')}"


def dtm_config_to_md(tape: Tape, rule_str: str) -> str:
    header_str = rule_to_md(rule_str)
    tape_str = tape_to_md(tape)

    return f"{header_str}\n<pre>\n{tape_str}</pre>\n---\n"


def mtm_config_to_md(tapes: List[Tape], rule_str: str) -> str:
    header_str = rule_to_md(rule_str)
    tapes_str = "".join(tape_to_md(tape, i) for i, tape in enumerate(tapes))

    return f"{header_str}\n<pre>\n{tapes_str}</pre>\n---\n"


def read_file(file_path: str) -> List[str]:
    result = []
    with open(file_path, "r") as in_file:
        for line in in_file:
            line = " ".join(line.split())

            if not line.strip() or line.startswith("#"):
                continue

            result.append(f"{line.strip()}")

    return result


def parse_direction(direction: str) -> Optional[Direction]:
    directions = {"L": -1, "R": 1, "S": 0}
    return Direction(directions.get(direction, 0))


def get_dtm_configuration(
    definition: List[str],
) -> Tuple[str, str, str, Set[str]]:
    config = list(takewhile(lambda l: l.strip() != "---", definition))[:-1]
    init, acc, rej = "", "", ""
    alphabet = set()

    for line in config:
        if line.startswith("init"):
            init = line.split()[1]
        elif line.startswith("acc"):
            acc = line.split()[1]
        elif line.startswith("rej"):
            rej = line.split()[1]
        elif line.startswith("alphabet"):
            alphabet = {*line.split()[1:]}

    return init, acc, rej, alphabet or set()


def get_mtm_configuration(
    definition: List[str],
) -> Tuple[str, str, str, Set[str], int]:
    config = next(
        (
            idk
            for idk in list(takewhile(lambda l: l.strip() != "---", definition))
            if idk.strip().startswith("tapes")
        )
    )
    tape_count = int(config.split()[1])

    return (*get_dtm_configuration(definition), tape_count)


def get_dtm_transition_function(definition: List[str]):
    function_lines = list(dropwhile(lambda l: l != "---", definition))[1:]
    function = {}

    for rule in function_lines:
        left, right = rule.split("->")
        curr_state, read = left.split()
        next_state, write, direction = right.split()

        if not function.get(curr_state):
            function[curr_state] = {}

        read = read if read != "_" else ""
        write = write if write != "_" else ""
        function[curr_state][read] = (next_state, write, parse_direction(direction))

    return function


def validate_dtm_configuration(definition: List[str]) -> Optional[str]:
    if all(l.strip() != "---" for l in definition):
        return "The divider is missing."

    config = list(takewhile(lambda l: l != "---", definition))

    if not any((l.startswith("init") for l in config)):
        return "Specifying the initial state is mandatory."

    init_line = next((l for l in config if l.startswith("init")), None)
    if init_line and len(init_line.split()) != 2:
        return "Invalid initial state."

    acc_line = next((l for l in config if l.startswith("acc")), None)
    if acc_line and len(acc_line.split()) != 2:
        return "Invalid accepting state."

    rej_line = next((l for l in config if l.startswith("rej")), None)
    if rej_line and len(rej_line.split()) != 2:
        return "Invalid rejecting state."

    alphabet_line = next((l for l in config if l.startswith("alphabet")), None)
    if not alphabet_line:
        return "The alphabet definition is missing."

    if len(alphabet_line.split()) < 2 or any(len(s) > 1 for s in alphabet_line.split()[1:]):
        return "Invalid alphabet."


def validate_mtm_configuration(definition: List[str]) -> Optional[str]:
    err = validate_dtm_configuration(definition)
    if err:
        return err

    tapes_line = [
        l for l in definition if re.match(r"tapes ([2-9]|[1-9]\d+)", l.strip())
    ]
    if len(tapes_line) > 1:
        return "Duplicate definition of tape count."

    if not tapes_line:
        return "Missing or invalid definition of tape count."


def validate_dtm_transitions(definition: List[str]) -> Optional[str]:
    get_part = lambda l, i: l.split("->")[i]
    
    # checks the length of arguments on the left side
    valid_current = lambda l: len(get_part(l, 0).split()) == 2
    
    # checks the length of the read symbol
    valid_read = lambda l: len(get_part(l, 0).split()[1]) == 1
    
    # checks the length of arguments on the right side
    valid_next = lambda l: len(get_part(l, 1).split()) == 3
    
    # checks the length of the write symbol
    valid_write = lambda l: len(get_part(l, 1).split()[1]) == 1
    
    # checks the direction
    valid_dir = lambda l: get_part(l, 1).split()[-1] in ["L", "R", "S"]

    return validate_transitions(
        definition,
        valid_current,
        valid_read,
        valid_next,
        valid_write,
        valid_dir,
    )


def validate_mtm_transitions(definition: List[str]) -> Optional[str]:
    get_part = lambda l, i: l.split("->")[i]
    # checks the length of arguments on the left side
    valid_current = lambda l: re.match(r"^\w+ \([^)]+\)", get_part(l, 0)) is not None
    
    # checks the length of the read symbols
    valid_read = (
        lambda l: re.match(r"(. )+.", get_part(l, 0).split(" (")[1][:-1]) is not None
    )
    # checks the length of arguments on the right side
    valid_next = lambda l: len(list(filter(None, get_part(l, 1).split(" (")))) == 3
    
    # checks the length of the write symbols
    valid_write = (
        lambda l: bool(re.match(r"^(. )+.$", get_part(l, 1).split(" (")[1][:-1])) == 1
    )
    
    # checks the direction
    valid_dir = lambda l: set(get_part(l, 1).split(" (")[-1][:-1].split()).issubset(
        ["L", "R", "S"]
    )

    return validate_transitions(
        definition,
        valid_current,
        valid_read,
        valid_next,
        valid_write,
        valid_dir,
    )


def validate_transitions(
    definition: List[str],
    valid_current: Callable[[str], bool],
    valid_read: Callable[[str], bool],
    valid_next: Callable[[str], bool],
    valid_write: Callable[[str], bool],
    valid_dir: Callable[[str], bool],
) -> Optional[str]:

    for line in list(dropwhile(lambda l: l != "---", definition))[1:]:
        if "->" not in line:
            return f"Missing arrow in rule:\n{line}."

        if not valid_current(line):
            return f"Invalid combination of state and read symbol in rule:\n{line}"

        if not valid_read(line):
            return f"Invalid read symbol length (> 1) in rule:\n{line}"

        if not valid_next(line):
            return (
                f"The next state, write symbol or direction is missing in rule:\n{line}"
            )

        if not valid_write(line):
            return f"Invalid write symbol length (> 1) in rule:\n{line}"

        if not valid_dir(line):
            return f"Invalid direction in rule:\n{line}"


def get_mtm_transition_function(definition: List[str]):
    function_lines = list(dropwhile(lambda l: l != "---", definition))[1:]
    function = {}

    def clean(group: str) -> List[str]:
        """Removes brackets and spaces + returns as list."""
        return re.sub(r"[()]", "", group).split()

    def underscore_to_space(group: List[str]) -> List[str]:
        return [s if s != "_" else "" for s in group]

    for rule in function_lines:
        left, right = rule.split("->")
        curr_state, reads = left.split(" ", 1)
        next_state, writes, directions = right.split("(")

        if not function.get(curr_state):
            function[curr_state] = {}

        function[curr_state][tuple(underscore_to_space(clean(reads)))] = (
            next_state.strip(),
            tuple(underscore_to_space(clean(writes))),
            tuple(parse_direction(d) for d in clean(directions)),
        )

    return function


if __name__ == "__main__":
    pass



from typing import Set


class BaseFiniteAutomaton:
    """
    Represents an abstract Finite Automaton class. This class cannot be instantiated.
    """

    def __new__(cls, *args, **kwargs):
        if cls is BaseFiniteAutomaton:
            raise TypeError("Only NFA and DFA can be instantiated!")

        return object.__new__(cls)

    def __init__(
        self,
        states: Set[str] = None,
        alphabet: Set[str] = None,
        initial_state: str = None,
        final_states: Set[str] = None,
    ) -> None:
        self.states = states if states is not None else set()
        self.alphabet = alphabet if alphabet is not None else set()
        self.initial_state = initial_state
        self.final_states = final_states if final_states is not None else set()

    def __repr__(self) -> str:
        alphabet_str = f"alphabet: {','.join(sorted(self.alphabet))}"
        states_str = f"states: {','.join(sorted(self.states))}"
        initial_str = f"initial state: {self.initial_state}"
        final_states_str = f"final states: {','.join(sorted(self.final_states))}"

        return f"{alphabet_str}\n{states_str}\n{initial_str}\n{final_states_str}\n"

    def __repr_row_prefix__(self, state: str):
        if state == self.initial_state:
            return "<-> " if state in self.final_states else "--> "

        return "<-- " if state in self.final_states else "    "

    def add_state(self, state: str, is_final: bool = False) -> bool:
        if state in self.states:
            return False

        if is_final:
            self.final_states.add(state)

        self.states.add(state)

        return True

    def remove_state(self, state) -> bool:
        if state not in self.states:
            return False

        if state in self.final_states:
            self.final_states.remove(state)

        self.states.difference_update({state})

        return True

    def is_valid(self) -> bool:
        return (
            self.states
            and self.initial_state in self.states
            and self.final_states.issubset(self.states)
        )

    def complement(self) -> None:
        """
        Complements the automaton. (Final states will become non-final and vice-versa).
        """
        self.final_states = self.states - self.final_states


if __name__ == "__main__":
    pass



from sys import path
from hypothesis import given, assume
from hypothesis.strategies import (
    integers,
    sets,
    characters,
    composite,
    DrawFn,
    lists,
    sampled_from,
)
from re import match
from tests.generation import r_nfa
from random import choice

path.append("../src/ib110hw")

from automaton.nfa import NFA


@composite
def r_test_nfa(
    draw: DrawFn,
    min_deg=integers(min_value=1, max_value=10),
    max_deg=integers(min_value=1, max_value=10),
    min_states=integers(min_value=2, max_value=10),
    max_states=integers(min_value=2, max_value=10),
    min_fin_states=integers(min_value=1, max_value=10),
    max_fin_states=integers(min_value=2, max_value=10),
    alphabet=sets(characters(), min_size=2, max_size=20),
):
    max_deg = draw(max_deg)
    min_deg = min(draw(min_deg), max_deg)
    max_states = draw(max_states)
    min_states = min(draw(min_states), max_states)
    max_fin_states = draw(max_fin_states)
    min_fin_states = min(draw(min_fin_states), max_fin_states)

    r_automaton = r_nfa(
        min_deg,
        max_deg,
        min_states,
        max_states,
        min_fin_states,
        max_fin_states,
        draw(alphabet),
    )

    diff = r_automaton.states.difference(r_automaton.final_states)
    assume(diff and diff != {r_automaton.initial_state})

    return r_automaton


@given(r_test_nfa())
def test_add_state(automaton: NFA) -> None:
    assert automaton.add_state("test_state_1")

    # cannot add the same state twice
    assert not automaton.add_state("test_state_1")
    assert not automaton.add_state("test_state_1", True)
    assert "test_state_1" in automaton.states
    assert "test_state_1" not in automaton.final_states
    assert "test_state_1" not in automaton.transitions

    assert automaton.add_state("test_state_2", True)
    assert not automaton.add_state("test_state_2")
    assert "test_state_2" in automaton.states
    assert "test_state_2" in automaton.final_states


@given(r_test_nfa())
def test_remove_state(automaton: NFA) -> None:
    state_to_remove = choice(
        [s for s in automaton.states if s not in automaton.final_states]
    )

    assert automaton.remove_state(state_to_remove)
    # cannot remove the same state twice
    assert not automaton.remove_state(state_to_remove)
    assert state_to_remove not in automaton.states
    assert not automaton.remove_state("not_existent")

    state_to_remove = choice(list(automaton.final_states))

    assert automaton.remove_state(state_to_remove)
    assert state_to_remove not in automaton.states
    assert state_to_remove not in automaton.final_states
    assert not automaton.transitions.get(state_to_remove, None)

    automaton.add_state("test_final_state", True)
    # just in case the final states was empty
    automaton.add_state("filler_state", True)

    assert automaton.remove_state("test_final_state")
    assert "test_final_state" not in automaton.final_states


@given(r_test_nfa())
def test_add_transition(automaton: NFA) -> None:
    assert automaton.add_transition("test_state", "test_state", "a")
    assert "test_state" in automaton.transitions.keys()
    assert "a" in automaton.transitions["test_state"]
    assert {"test_state"} == automaton.transitions["test_state"]["a"]
    assert not automaton.add_transition("test_state", "test_state", "a")

    automaton.add_state("test_second_state")
    assert automaton.add_transition("test_state", "test_second_state", "a")
    assert "test_second_state" not in automaton.transitions.keys()
    assert {"test_state", "test_second_state"} == automaton.get_transition(
        "test_state", "a"
    )
    assert not automaton.add_transition("test_state", "test_second_state", "a")


@given(r_test_nfa())
def test_set_transition(automaton: NFA) -> None:
    automaton.set_transition("s_from", {"s_to1", "s_to2"}, "symbol")
    assert automaton.get_transition("s_from", "symbol") == {"s_to1", "s_to2"}

    automaton.set_transition("s_from", {"s_to2"}, "symbol2")
    assert automaton.get_transition("s_from", "symbol2") == {"s_to2"}
    assert automaton.get_transition("s_from", "symbol") == {"s_to1", "s_to2"}

    automaton.set_transition("s_from", {"s_to2"}, "symbol")
    assert automaton.get_transition("s_from", "symbol") == {"s_to2"}


@given(r_test_nfa())
def test_remove_transition(automaton: NFA) -> None:
    automaton.add_state("test_state")
    automaton.alphabet.add("a")
    automaton.add_transition("test_state", "test_state", "a")

    assert not automaton.remove_transition(
        "test_state", automaton.alphabet.difference(["a"]).pop(), "test_state"
    )
    assert not automaton.remove_transition("not_exists", "test_state", "a")
    assert automaton.remove_transition("test_state", "test_state", "a")
    assert not automaton.get_transition("test_state", "a")
    assert not automaton.remove_transition("test_state", "test_state", "a")

    automaton.add_state("test_second_state")
    automaton.add_transition("test_state", "test_state", "a"),
    automaton.add_transition("test_state", "test_second_state", "a")
    assert automaton.remove_transition("test_state", "test_state", "a")
    assert automaton.get_transition("test_state", "a") == {"test_second_state"}


@composite
def r_input_string_acc(draw: DrawFn):
    length = draw(integers(min_value=1, max_value=50))
    insert_idx = draw(integers(min_value=0, max_value=length))
    substr = draw(sampled_from(["cba", "cbaa", "ca", "cb"]))
    fill_str = "".join(
        draw(lists(sampled_from("abc"), min_size=length, max_size=length))
    )

    return fill_str[:insert_idx] + substr + fill_str[insert_idx:]


@composite
def r_input_string_rej(draw: DrawFn):
    length = draw(integers(min_value=2, max_value=50))
    result = "".join(draw(lists(sampled_from("abc"), min_size=length, max_size=length)))

    while match(r".*(cba|cbaa|ca|cb).*", result):
        for acc_str in ["cba", "cbaa", "ca", "cb"]:
            result = result.replace(acc_str, "")

    assume(len(result) > 1)

    return result


@given(r_input_string_acc(), r_input_string_rej())
def test_is_accepted(acc_str: str, rej_str: str) -> None:
    automaton: NFA = NFA(
        states={f"s{i}" for i in range(10)},
        alphabet={"a", "b", "c"},
        initial_state="s0",
        final_states={"s3", "s5", "s7", "s9"},
        transitions={
            "s0": {"a": {"s0"}, "b": {"s0"}, "c": {"s0", "s1", "s6", "s8"}},
            "s1": {"b": {"s2"}},
            "s2": {"a": {"s3", "s4"}},
            "s3": {"a": {"s3"}, "b": {"s3"}, "c": {"s3"}},
            "s4": {"a": {"s5"}},
            "s5": {"a": {"s5"}, "b": {"s5"}, "c": {"s5"}},
            "s6": {"a": {"s7"}},
            "s7": {"a": {"s7"}, "b": {"s7"}, "c": {"s7"}},
            "s8": {"b": {"s9"}},
            "s9": {"a": {"s9"}, "b": {"s9"}, "c": {"s9"}},
        },
    )

    assert automaton.is_accepted(acc_str), f"Expected '{acc_str}' to be accepted."
    assert not automaton.is_accepted(rej_str), f"Expected '{rej_str}' to be rejected."


def test_is_valid() -> None:
    automaton: NFA = NFA(
        states={"s0", "s1", "s2"},
        alphabet={"a", "b", "c", ""},
        initial_state="s0",
        final_states={"s1", "s2"},
        transitions={
            "s0": {"a": {"s0", "s1"}, "b": {"s2"}},
            "s1": {"a": {"s1"}, "b": {"s0", "s1", "s2"}, "c": {"s1"}},
            "s2": {"a": {"s0", "s1"}, "": {"s1"}},
        },
    )

    assert automaton.is_valid()

    # invalid initial state
    automaton.initial_state = "s3"
    assert not automaton.is_valid()

    automaton.initial_state = "s0"
    assert automaton.is_valid()

    # invalid state in the transition function
    automaton.add_transition("s0", "s3", "b")
    assert not automaton.is_valid()

    automaton.remove_transition("s0", "s3", "b")
    assert automaton.is_valid()

    # final states set is not subset of the states set
    automaton.final_states.update({"s3"})
    assert not automaton.is_valid()

    automaton.final_states.remove("s3")
    assert automaton.is_valid()

    # invalid character in the transition function
    automaton.add_transition("s0", "s2", "x")
    assert not automaton.is_valid()

    automaton.remove_transition("s0", "s2", "x")
    assert automaton.is_valid()

    # empty states set
    automaton.states = set()
    assert not automaton.is_valid()
